package org.openxma.dsl.generator.ant;

import java.io.File;
import java.io.IOException;
import java.net.URL;
import java.net.URLClassLoader;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Set;

import org.apache.tools.ant.BuildException;
import org.apache.tools.ant.DirectoryScanner;
import org.apache.tools.ant.types.FileSet;
import org.apache.tools.ant.types.Path;
import org.apache.tools.ant.types.resources.FileResource;
import org.eclipse.emf.mwe.core.ant.WorkflowAntTask;
import org.eclipse.emf.mwe.internal.core.ant.Parameter;
import org.hibernate.MappingException;
import org.hibernate.cfg.Configuration;
import org.hibernate.dialect.Dialect;
import org.hibernate.jdbc.util.DDLFormatterImpl;
import org.openxma.dsl.generator.helper.FileUtil;

/**
 * Ant task to used to generate table creation scripts based on the configured
 * set of mapping files.<br>
 * Pls. note that this task must be executed within some 'process-classes' (i.e.
 * after compile-classes) phase of <br>
 * your build because it requires access to the compiled hibernate classes and
 * mapping files. <h2>Usage:</h2> <h3>Taskdef:</h3>
 * 
 * <pre>
 * &lt;taskdef name="xmaDdlGen"
 *          classname="org.openxma.dsl.generator.ant.DdlGeneratorTask"
 *          classpathref="xmaDdlGen.classpath" /&gt;
 * </pre>
 * 
 * <h3>Classpath:</h3> The classpath must contain<br>
 * <ul>
 * <li>The project output directory (destdir attribute from javac ant-task),
 * containing the compiled entities and mapping file(s)</li>
 * <li>EMF runtime jars (<code>ecore-2.5.0.jar</code>,
 * <code>common-2.5.0.jar</code> and <code>emf-xmi-2.5.0.jar</code>)</li>
 * <li>MWE runtime jars (<code>emf-mwe-core-1.0.1.jar</code>,
 * <code>emf-mwe-utils-1.0.1.jar</code> and
 * <code>emf-mwe2-runtime-1.0.1.jar</code>)</li>
 * <li>Xpand runtime jars (<code>m2t-xpand-xpand-1.0.1.jar</code>,
 * <code>m2t-xpand-xtend-1.0.1.jar</code> and
 * <code>m2t-xpand-xtend-typesystem-emf-1.0.1.jar</code>)</li>
 * <li>Xtext and Antlr runtime jars (<code>xtext-core-1.0.1.jar</code>,
 * <code>xtext-util-1.0.1.jar</code> and <code>antlr-3.0.1.jar</code>)</li>
 * <li>Google collect and inject (<code>collect-0.8.0.jar</code> and
 * <code>guice-2.0.jar</code>)</li>
 * <li>Apache libraries (<code>log4j-1.2.16.jar</code>,
 * <code>commons-collections-3.2.1.jar</code>,
 * <code>commons-logging-1.1.1.jar</code> and <code>commons-cli-1.2.jar</code>)</li>
 * <li>Hibernate jars and dependencies (<code>hibernate-3.6.8.Final.jar</code> and
 * <code>dom4j-1.6.1.jar</code>)</li>
 * <li>openXMA jars (<code>dsl-ddl-3.6.2.jar</code>,
 * <code>dsl-platform-3.6.2.jar</code> and
 * <code>dsl-generator-3.6.2.jar</code>)</li>
 * <li>other (<code>icu-4.0.1.jar</code>, spring-3.1.0 jars)<br>
 * </li>
 * </ul>
 * 
 * 
 * <h3>Usage with with all possible settings:</h3>
 * 
 * <pre>
 * &lt;xmaDdlGen ddldir="doc/database" ddlFile="create_tables.ddl"&gt;
 *         classdir="classes" classpathref="ddlGenTask.classpath"
 *         hibernateDialect="org.hibernate.dialect.Oracle10gDialect"
 *     &lt;hibernateMappings dir="src-gen"&gt;
 *         &lt;include name="**&#47;*.hbm.xml"/&gt;
 *         &lt;exclude name="**&#47;MySqlTable.hbm.xml"/&gt;
 *     &lt;/hibernateMappings&gt;
 * &lt;/xmaDdlGen&gt;
 * </pre>
 * 
 */
public class DdlGeneratorTask extends WorkflowAntTask {
    private static final String DEFAULT_HIBERNATE_DIALECT = "org.openxma.dsl.platform.hibernate.dialect.ConfigurableOracle10gDialect";
    private FileSet hibernateMappings;
    private String hibernateDialect = DEFAULT_HIBERNATE_DIALECT;
    private File tmpDir;
    private String ddlFile = "create_tables.ddl";
    private File ddlDir;
    private File classDir;
    private boolean skipMappingErrors;

    public void setSkipMappingErrors(boolean skipMappingErrors) {
        this.skipMappingErrors = skipMappingErrors;
    }

    public void addHibernateMappings(FileSet hibernateMappings) {
        this.hibernateMappings = hibernateMappings;
    }

    public String getHibernateDialect() {
        return hibernateDialect;
    }

    public void setHibernateDialect(String hibernateDialect) {
        this.hibernateDialect = hibernateDialect;
    }

    public File getTmpDir() {
        return tmpDir;
    }

    public void setTmpDir(File tmpDirectory) {
        this.tmpDir = tmpDirectory;
    }

    public String getDdlFile() {
        return ddlFile;
    }

    public void setDdlFile(String ddlFileName) {
        this.ddlFile = ddlFileName;
    }

    public File getDdlDir() {
        return ddlDir;
    }

    public void setDdlDir(File ddlDirectory) {
        this.ddlDir = ddlDirectory;
    }

    public File getClassDir() {
        return classDir;
    }

    public void setClassDir(File classDir) {
        this.classDir = classDir;
    }

    @Override
    public void execute() throws BuildException {
        assertConfiguration();
        ClassLoader contextClassLoader = Thread.currentThread().getContextClassLoader();
        try {
            initTCCL();
            String tableCreationScript = createTableCreationScript();
            File tmpDirectory = saveToTmpDirectory(tableCreationScript);
            setFile("workflow/DdlGenerator.mwe");
            createParam("ddlFileName", ddlFile);
            createParam("ddlDirectory", ddlDir.toString());
            createParam("modelFolder", tmpDirectory.toString());
            super.execute();
        } finally {
            resetTCCL(contextClassLoader);
        }
    }

    private void assertConfiguration() {
        if (ddlDir == null) {
            throw new BuildException("Missing attribute 'ddlDir' or specified directory does not exist.");
        }
        if (hibernateMappings == null) {
            throw new BuildException("Missing attribute 'hibernateMappings'.");
        }
        if (!ddlDir.exists()) {
            ddlDir.mkdirs();
        }
    }

    private void createParam(String name, String value) {
        Parameter parameter = (Parameter) createParam();
        parameter.setName(name);
        parameter.setValue(value);
    }

    private File saveToTmpDirectory(String tableCreationScript) throws BuildException {
        try {
            return FileUtil.writeStringToFile(tableCreationScript,
                    new File(FileUtil.createTempDirectory(), "_" + ddlFile)).getParentFile();
        } catch (IOException e) {
            throw new BuildException(e.getMessage(), e);
        }
    }

    private void initTCCL() {
        try {
            List<URL> classPathUrls = new ArrayList<URL>(Collections.singletonList(getClassDir().toURI().toURL()));
            Path classpath = getCommandLine().getClasspath();
            @SuppressWarnings("unchecked")
            Iterator<FileResource> iterator = classpath.iterator();
            while (iterator.hasNext()) {
                FileResource next = iterator.next();
                classPathUrls.add(next.getFile().toURI().toURL());
            }
            Thread.currentThread().setContextClassLoader(
                    URLClassLoader.newInstance(classPathUrls.toArray(new URL[] {}), getClass().getClassLoader()));
        } catch (Exception e) {
            throw new BuildException(e.getMessage(), e);
        }
    }

    private void resetTCCL(ClassLoader contextClassLoader) {
        Thread.currentThread().setContextClassLoader(contextClassLoader);
    }

    private String createTableCreationScript() {
        DirectoryScanner directoryScanner = this.hibernateMappings.getDirectoryScanner(getProject());
        Set<String> processedMappingFiles = new HashSet<String>();
        String[] includedFiles = directoryScanner.getIncludedFiles();
        Configuration configuration = new Configuration();
        for (String includedFile : includedFiles) {
            File file = new File(getProject().getBaseDir(), includedFile);
            if (processedMappingFiles.add(file.getName())) {
                try {
                    configuration.addFile(file);
                } catch (MappingException mappingException) {
                    if (!skipMappingErrors) {
                        throw new BuildException(mappingException.getMessage(), mappingException);
                    }
                }
            }
        }
        String[] creationScript = configuration.generateSchemaCreationScript(createHibernateDialect());
        StringBuilder ddlBuilder = new StringBuilder();
        DDLFormatterImpl ddlFormatter = new DDLFormatterImpl();
        for (String ddl : creationScript) {
            String string = ddlFormatter.format(ddl);            
            ddlBuilder.append(string);
            ddlBuilder.append(";");
        }
        return ddlBuilder.toString();
    }

    private Dialect createHibernateDialect() {
        Dialect result = null;
        try {
            Class<?> hibernateDialect = (getHibernateDialect() != null ? getClass().getClassLoader().loadClass(
                    getHibernateDialect()) : getClass().getClassLoader().loadClass(DEFAULT_HIBERNATE_DIALECT));
            result = (Dialect) hibernateDialect.newInstance();
            log("used hibernateDialect is '" + hibernateDialect + "'");
        } catch (Exception e) {
            throw new BuildException(e.getMessage(), e);
        }
        return result;
    }

}
