package org.openxma.dsl.generator.impl;

import static org.openxma.dsl.generator.GeneratorConstants.DOMAIN_MODEL_CONTEXT_IMPL_PACKAGE;
import static org.openxma.dsl.generator.GeneratorConstants.DOMAIN_MODEL_CONTEXT_PACKAGE;
import static org.openxma.dsl.generator.GeneratorConstants.DOMAIN_MODEL_DAO_IMPL_PACKAGE;
import static org.openxma.dsl.generator.GeneratorConstants.DOMAIN_MODEL_DAO_PACKAGE;
import static org.openxma.dsl.generator.GeneratorConstants.DOMAIN_MODEL_DTO_PACKAGE;
import static org.openxma.dsl.generator.GeneratorConstants.DOMAIN_MODEL_ENUM_PACKAGE;
import static org.openxma.dsl.generator.GeneratorConstants.DOMAIN_MODEL_IMPL_PACKAGE;
import static org.openxma.dsl.generator.GeneratorConstants.DOMAIN_MODEL_PACKAGE;
import static org.openxma.dsl.generator.GeneratorConstants.DOMAIN_MODEL_SERVICE_IMPL_PACKAGE;
import static org.openxma.dsl.generator.GeneratorConstants.DOMAIN_MODEL_SERVICE_PACKAGE;
import static org.openxma.dsl.generator.GeneratorConstants.DOMAIN_MODEL_TYPE_PACKAGE;

import org.openxma.dsl.common.Strings;
import org.openxma.dsl.core.model.EnumType;
import org.openxma.dsl.core.model.Model;
import org.openxma.dsl.core.model.ModelElement;
import org.openxma.dsl.core.model.SimpleType;
import org.openxma.dsl.dom.model.ApplicationSession;
import org.openxma.dsl.dom.model.Attribute;
import org.openxma.dsl.dom.model.Dao;
import org.openxma.dsl.dom.model.DataView;
import org.openxma.dsl.dom.model.Entity;
import org.openxma.dsl.dom.model.Service;
import org.openxma.dsl.generator.GeneratorConfiguration;
import org.openxma.dsl.generator.NamingStrategy;

/**
 * Default implementation of interface <code>NamingStrategy</code>.
 */
public class DefaultNamingStrategy implements NamingStrategy {
    public static final String KEY_COLUMN_PREFIX = "ID_";

	public String getTableName(Dao dao) {
		return Strings.toUnderScoreString(dao.getEntity().getName()).toUpperCase();
	}

	public String getQualifier(Dao dao) {
		return Strings.removeVowelsAndUnderScore(dao.getEntity().getName()).toUpperCase();
	}

	public String getColumnName(Entity entity, Attribute attribute) {
		String columnName = null;
		if (attribute.isIdentifier()) {
			columnName = Strings.KEY_COLUMN_PREFIX + Strings.toColumnName(entity.getName());
		} else if (attribute.isReference()) {
			columnName = getForeignKeyColumnName(entity, attribute);
		} else if (attribute.isHstore()) {
		    columnName = Strings.toColumnName(attribute.getHstoreColumn());
		} else {		
			columnName = Strings.toColumnName(attribute.getName());
		}
		return columnName;
	}

	public String getColumnName(Entity entity, Attribute feature, Attribute nestedFeature) {
		String pathAbbreviation = Strings.removeVowelsAndUnderScore(feature.getName());
		String columnName = Strings.toColumnName(pathAbbreviation + "_" + nestedFeature.getName());
		return columnName;
	}

	public String getForeignKeyColumnName(Entity entity, Attribute reference) {
		String columnName = null;
		if (reference.isMany()) {
			columnName = Strings.toColumnName(entity.getName()).toUpperCase();
		} else {
			columnName = Strings.toColumnName(reference.getName()).toUpperCase();
		}
		if (!columnName.startsWith(Strings.KEY_COLUMN_PREFIX)) {
			columnName = Strings.KEY_COLUMN_PREFIX + columnName;
		}
		return columnName;

	}

	public String getPackageName(ModelElement modelElement) {
		if (modelElement instanceof Entity) {
			return getInterfacePackageName(modelElement) + "." + GeneratorConfiguration.INSTANCE.getProperty(DOMAIN_MODEL_IMPL_PACKAGE);
		} else if (modelElement instanceof Service) {
			return getInterfacePackageName(modelElement) + "." + GeneratorConfiguration.INSTANCE.getProperty(DOMAIN_MODEL_SERVICE_IMPL_PACKAGE);
		} else if (modelElement instanceof Dao) {
			return getInterfacePackageName(modelElement) + "." + GeneratorConfiguration.INSTANCE.getProperty(DOMAIN_MODEL_DAO_IMPL_PACKAGE);
		} else if (modelElement instanceof ApplicationSession) {
			return getInterfacePackageName(modelElement) + "." + GeneratorConfiguration.INSTANCE.getProperty(DOMAIN_MODEL_CONTEXT_IMPL_PACKAGE);
		} else if (modelElement instanceof DataView) {
			return getPackageNameFromPreferences((ModelElement) modelElement, DOMAIN_MODEL_DTO_PACKAGE);
		} else if (modelElement instanceof EnumType) {
			return getPackageNameFromPreferences((ModelElement) modelElement, DOMAIN_MODEL_ENUM_PACKAGE);
		} else if (modelElement instanceof SimpleType) {
			return getPackageNameFromPreferences((ModelElement) modelElement, DOMAIN_MODEL_TYPE_PACKAGE);
		} 
		throw new IllegalStateException("type '" + modelElement + "' is not expected");
	}

	public String getInterfacePackageName(ModelElement modelElement) {
		if (modelElement instanceof Entity) {
			return getPackageNameFromPreferences((ModelElement) modelElement, DOMAIN_MODEL_PACKAGE);
		} else if (modelElement instanceof Service) {
			return getPackageNameFromPreferences((ModelElement) modelElement, DOMAIN_MODEL_SERVICE_PACKAGE);
		} else if (modelElement instanceof Dao) {
			return getPackageNameFromPreferences((ModelElement) modelElement, DOMAIN_MODEL_DAO_PACKAGE);
		} else if (modelElement instanceof ApplicationSession) {
			return getPackageNameFromPreferences((ModelElement) modelElement, DOMAIN_MODEL_CONTEXT_PACKAGE);
		}
		throw new IllegalStateException("type '" + modelElement + "' is not expected");
	}

	private static String getPackageNameFromPreferences(ModelElement modelObject, String preferenceKey) {
		String modelPackageName = ((Model)modelObject.eContainer()).getName();
		String defaultPackageName = GeneratorConfiguration.INSTANCE.getProperty(preferenceKey);
		return internalGetPackage(modelPackageName, defaultPackageName);
	}

	private static String internalGetPackage(String modelPackageName, String defaultPackageName) {
		return modelPackageName.endsWith(defaultPackageName) ? modelPackageName : modelPackageName + "."
				+ defaultPackageName;
	}

}
